

//---------------------------------
//           Array 
//---------------------------------


template<class T>
Array<T>::Array(int n):size(n)
   {   
   int i;
   
   #ifdef ARRAY_ERR_CHK   
      if(!size) error(1);
   #endif
   
   m=new T*[size];

   #ifdef ARRAY_ERR_CHK   
      if(!m) error(2);
   #endif

   m--;
      
   refcnt=new int(1);
   }

template<class T>
Array<T>::Array(const Array<T> &a)
   {
   int i;
   
   refcnt=a.refcnt;
   (*refcnt)=(*refcnt)>=0 ? (*refcnt)+1 : 1;
   size=a.size;
   m=a.m;
   }

template<class T>
Array<T>& Array<T>::operator =(Array<T> &a)
   {
   int i;
   
   #ifdef ARRAY_ERR_CHK   
      if(size!=a.size) error(3);
   #endif

   if(a.m != m)
      {
      for(i=1;i<=size;i++)
         (*this)(i) = a(i);
      
      a.declone();  
      }
      
   return *this;  
   }

template<class T>
Array<T>& Array<T>::operator =(T k)
   {
   int i;
   
   for(i=1;i<=size;i++)
      (*this)(i)=k;
   
   return *this;
   }

template<class T>
Array<T>& Array<T>::operator -()
   {
   Array<T>* b=clone();
   int i;
   
   for(i=1;i<=size;i++)
      (*b)(i) = -(*this)(i);

   *(b->refcnt)=-1;
   return *b; 
   }

template<class T>
Array<T>& operator +(Array<T> &a, Array<T> &b)
   {
   #ifdef ARRAY_ERR_CHK   
      if(b.size!=a.size) error(4);
   #endif

   Array<T>* c=a.clone();
   int i;
   
   for(i=1;i<=c->size;i++)
      (*c)(i)=a(i)+b(i);
   
   a.declone();
   b.declone();
   
   *(c->refcnt)=-1;
   return *c;
   }

template<class T>
Array<T>& operator -(Array<T> &a, Array<T> &b)
   {
   #ifdef ARRAY_ERR_CHK   
      if(b.size!=a.size) error(5);
   #endif

   Array<T>* c=a.clone();
   int i;
   
   for(i=1;i<=c->size;i++)
      (*c)(i)=a(i)-b(i);
   
   a.declone();
   b.declone();
   
   *(c->refcnt)=-1;
   return *c;
   }

template<class T>
void Array<T>::operator +=(const Array<T> &a)
   {
   #ifdef ARRAY_ERR_CHK   
      if(size!=a.size) error(4);
   #endif

   int i;
   
   for(i=1;i<=size;i++)
      (*this)(i)+=a(i);
   }

template<class T>
void Array<T>::operator -=(const Array<T> &a)
   {
   #ifdef ARRAY_ERR_CHK   
      if(size!=a.size) error(5);
   #endif

   int i;
   
   for(i=1;i<=size;i++)
      (*this)(i)-=a(i);
   }
   
template<class T>
void Array<T>::operator +=(T k)
   {
   int i;
   
   for(i=1;i<=size;i++)
      (*this)(i)+=k;
   }

template<class T>
void Array<T>::operator -=(T k)
   {
   int i;
   
   for(i=1;i<=size;i++)
      (*this)(i)-=k;
   }

template<class T>
Array<T>& operator *(Array<T> &a, Array<T> &b)
   {
   #ifdef ARRAY_ERR_CHK   
      if(b.size!=a.size) error(6);
   #endif

   Array<T>* c=a.clone();
   int i;
   
   for(i=1;i<=c->size;i++)
      (*c)(i)=a(i)*b(i);
   
   a.declone();
   b.declone();
   *(c->refcnt)=-1;

   return *c;
   }

template<class T>
Array<T>& operator /(Array<T> &a, Array<T> &b)
   {
   #ifdef ARRAY_ERR_CHK   
      if(b.size!=a.size) error(7);
   #endif

   Array<T>* c=a.clone();
   int i;
   
   for(i=1;i<=c->size;i++)
      (*c)(i)=a(i)/b(i);
   
   a.declone();
   b.declone();
   *(c->refcnt)=-1;

   return *c;
   }
   
template<class T>
Array<T>& operator *(Array<T> &a,T k)
   {
   Array<T>* b=a.clone();
   int i;
   
   for(i=1;i<=b->size;i++)
      (*b)(i)=a(i)*k;

   a.declone();
   *(b->refcnt)=-1;

   return *b;
   }

template<class T>
Array<T>& operator *(T k,Array<T> &a)
   {
   Array<T>* b=a.clone();
   int i;
   
   for(i=1;i<=b->size;i++)
      (*b)(i)=a(i)*k;

   a.declone();
   *(b->refcnt)=-1;

   return *b;
   }

template<class T>
void Array<T>::operator *=(T k)
   {
   int i;
   
   for(i=1;i<=size;i++)
      (*this)(i)*=k;
   }

template<class T>
Array<T>& Array<T>::operator <<(int di) const
   {
   Array<T>* a=clone();
   int i;
   
   di--;
   for(i=1;i<=size;i++)
      (*a)(i)=(*this)((i+di)%size+1);
   
   *(a->refcnt)=-1;
   return *a;
   }   
   
template<class T>
Array<T>& Array<T>::operator >>(int di) const
   {
   Array<T>* a=clone();
   int i;
   
   di++;
   for(i=1;i<=size;i++)
      (*a)(i)=(*this)((i-di+size)%size+1);
   
   *(a->refcnt)=-1;
   return *a;
   }   

template<class T>
void Array<T>::linear(T xa,T xb)
   {
   T dx=(xb-xa)/T(size-1);
   int i;
   
   (*this)(1)=xa;
   for(i=2;i<=size;i++)
      (*this)(i)=(*this)(i-1)+dx;      
   }
   
template<class T>
T Array<T>::maximum(int &imax)
   {
   int i;
   T maxi=(*this)(1);
   
   imax=1;
   for(i=2;i<=size;i++)
      if((*this)(i)>maxi)
         maxi=(*this)(imax=i);
   
   return maxi;
   }

template<class T>
T Array<T>::minimum(int &imin)
   {
   int i;
   T mini=(*this)(1);
   
   imin=1;
   for(i=2;i<=size;i++)
      if((*this)(i)<mini)
         mini=(*this)(imin=i);
   
   return mini;
   }



//---------------------------------
//           Matrix 
//---------------------------------


template<class T>
Matrix<T>::Matrix(int nr):Array<T>(1)
   {
   int i;
   
   nrow=nr; ncol=1; size=nr;
   
   #ifdef ARRAY_ERR_CHK   
      if(!size) error(8);
   #endif

   m[1]=new T[size];

   #ifdef ARRAY_ERR_CHK   
      if(!m[1]) error(9);
   #endif

   m[1]--;
   }

template<class T>
Matrix<T>::Matrix(int nc, int nr):Array<T>(nc)
   {
   int i;
   
   nrow=nr; ncol=nc; size=nrow*ncol;

   #ifdef ARRAY_ERR_CHK   
      if(!size) error(8);
   #endif
   
   m[1]=new T[size];

   #ifdef ARRAY_ERR_CHK   
      if(!m[1]) error(9);
   #endif

   m[1]--;
   
   for(i=2;i<=ncol;i++)
         m[i]=m[i-1]+nrow;
   }
   
template<class T>
Matrix<T>::Matrix(const Matrix<T>& a)
   {
   refcnt=a.refcnt;
   
   (*refcnt)+=((*refcnt)>0);
   nrow=a.nrow;
   ncol=a.ncol;
   size=a.size;
   m=a.m;
   }   

template<class T>
Matrix<T>::~Matrix()
   { if(refcnt) decrement(); }
   
template<class T>
void Matrix<T>::decrement()
   {
   if(refcnt)
      {
      (*refcnt)--;
      if(!(*refcnt))
         {
         delete [] (m[1]+1);
         delete [] (m+1);
         delete refcnt;
         }
      }   
   }

template<class T>
void Matrix<T>::declone()
   {
   if(*refcnt<0)
      {
      delete [] (m[1]+1);
      delete [] (m+1);
      delete refcnt;
      delete this;
      }
   }
           
template<class T>
void Matrix<T>::print()
   {
   int i,j;
   
   printf("\n");
   for(j=1;j<=nrow;j++)
      {
      for(i=1;i<=ncol;i++)
         printf("%10.4f",m[i][j]);
         printf("\n");
      }
      
   declone();   
   }   

template<class T>
size_t Matrix<T>::read(FILE *stream)
   {
   return fread(m[1],sizeof(T),size,stream);
   }

template<class T>
size_t Matrix<T>::write(FILE *stream)
   {
   return fwrite(m[1],sizeof(T),size,stream);
   }

template<class T>
Matrix<T>& Matrix<T>::operator =(Array<T> &a)
   {
   int i;
   
   #ifdef ARRAY_ERR_CHK   
      if(size!=a.n_elements()) error(10);
   #endif

   for(i=1;i<=size;i++)
      m[1][i] = a(i);
      
   a.declone();  
      
   return *this;  
   }

template<class T>
Matrix<T>& Matrix<T>::operator =(T k)
   {
   int i;
   
   for(i=1;i<=size;i++)
      m[1][i]=k;
   
   return *this;
   }

template<class T>
Partition<T> Matrix<T>::row(int j) const 
   {
   Partition<T> p(ncol);
   int i;
   
   p.ncol=ncol;
   p.nrow=1;
   for(i=1;i<=ncol;i++)
      p.m[i]=&(m[i][j])-1;
      
   return p;
   } 


template<class T>
Partition<T> Matrix<T>::col(int i) const
   {
   Partition<T> p(nrow);
   int j;
   
   p.ncol=1;
   p.nrow=nrow;
   for(j=1;j<=nrow;j++)
      p.m[j]=&(m[i][j])-1;
   
   return p;
   } 

template<class T>
Partition<T> Matrix<T>::operator ()(int i1,int i2,int j1,int j2) const
   {
   int pncol=(i2-i1+1),pnrow=(j2-j1+1),i,j,ip;
   Partition<T> p(pnrow*pncol);
   
   p.ncol=pncol;
   p.nrow=pnrow;
   ip=1;
   for(i=i1;i<=i2;i++)
      for(j=j1;j<=j2;j++,ip++)
         p.m[ip]=&(m[i][j])-1;
      
   return p;
   } 

template<class T>
void Matrix<T>::ref(const Matrix<T> &a)
   {
   decrement();
   refcnt=a.refcnt;
   
   (*refcnt)=(*refcnt)>=0 ? (*refcnt)+1 : 1;
   nrow=a.nrow;
   ncol=a.ncol;
   size=a.size;
   m=a.m;
   }   

template<class T>
void Matrix<T>::add(const Matrix<T> &a,const Matrix<T> &b)
   {
   #ifdef ARRAY_ERR_CHK   
      if(b.nrow!=a.nrow || b.ncol!=a.ncol || 
           nrow!=a.nrow || ncol!=a.ncol)
         error(11);
   #endif

   int i;
   
   for(i=1;i<=size;i++)
      m[1][i]=b.m[1][i]+a.m[1][i];
   }

template<class T>
void Matrix<T>::sub(const Matrix<T> &a,const Matrix<T> &b)
   {
   #ifdef ARRAY_ERR_CHK   
      if(b.nrow!=a.nrow || b.ncol!=a.ncol ||
           nrow!=a.nrow || ncol!=a.ncol)
         error(12);
   #endif

   int i;
   
   for(i=1;i<=size;i++)
      m[1][i]=b.m[1][i]-a.m[1][i];
   }


//---------------------------------
//           Partition 
//---------------------------------


template<class T>
Partition<T>::Partition(int nr):Array<T>(nr)
   {   
   nrow=nr; ncol=1;
   }

template<class T>
Partition<T>::Partition(const Partition<T> &a)
   {
   int i;
   
   refcnt=a.refcnt;
   (*refcnt)+=((*refcnt)>0);
   ncol=a.ncol;
   size=a.size;
   nrow=a.nrow;
   m=a.m;
   }

template<class T>
Partition<T>::Partition(const Matrix<T> &a):Array<T>(a.n_elements())
   {
   int i;
   
   nrow=a.n_rows(); 
   ncol=a.n_columns();
   size=a.n_elements();
   
   for(i=1;i<=size;i++)
      m[i]=&(a(i))-1;
   }

template<class T>
Partition<T>::~Partition()
   { decrement(); }
   
template<class T>
void Partition<T>::decrement()   
   {
   if(refcnt)
      { 
      (*refcnt)--;
      if(!(*refcnt))
         {
         delete [] (m+1);
         delete refcnt;
         }
      }
   }   
   
template<class T>
void Partition<T>::declone()
   {
   if(*refcnt<0)
      {
      delete [] (m+1);
      delete refcnt;
      delete this;
      }
   }

template<class T>
void Partition<T>::print()
   {
   int i,j;
   
   printf("\n");
   for(j=1;j<=nrow;j++)
      {
      for(i=1;i<=ncol;i++)
         printf("%10.4f",m[(i-1)*nrow+j][1]);
         printf("\n");
      }

   declone();   
   }   
      
template<class T>
Partition<T>& Partition<T>::operator =(Array<T> &a)
   {
   int i;
   
   #ifdef ARRAY_ERR_CHK   
      if(size!=a.n_elements()) error(13);
   #endif

   for(i=1;i<=size;i++)
      m[i][1] = a(i);
      
   a.declone();  
      
   return *this;  
   }

template<class T>
Partition<T>& Partition<T>::operator =(T k)
   {
   int i;
   
   for(i=1;i<=size;i++)
      m[i][1]=k;
   
   return *this;
   }

template<class T>
void Partition<T>::ref(const Partition<T> &a)
   {
   decrement();
   refcnt=a.refcnt;
   
   (*refcnt)=(*refcnt)>=0 ? (*refcnt)+1 : 1;
   nrow=a.nrow;
   ncol=a.ncol;
   size=a.size;
   m=a.m;
   }   
   
void error(int i)
   {
   char *msg;
   switch(i)
      {
      case 1:
         msg = "Array declaration error: size must be > 0";
         break;
      case 2:
         msg = "Array declaration: malloc error (out of memory ?)";
         break;
      case 3:
         msg = "error using Array assignment: different sizes";
         break;
      case 4:
         msg = "error using Array addition: different sizes";
         break;
      case 5:
         msg = "error using Array subtraction: different sizes";
         break;
      case 6:
         msg = "error using Array multiplication: different sizes";
         break;
      case 7:
         msg = "error using Array division: different sizes";
         break;
      case 8:
         msg = "Matrix declaration error: size must be > 0";
         break;
      case 9:
         msg = "Matrix declaration: malloc error (out of memory ?)";
         break;
      case 10:
         msg = "error using Matrix assignment: different sizes";
         break;
      case 11:
         msg = "error using Matrix addition: different shapes";
         break;
      case 12:
         msg = "error using Matrix subtraction: different shapes";
         break;
      case 13:
         msg = "error using Partition assignment: different sizes";
         break;
      default:
         msg = "class Array: fatal error";  
      }   
   printf("\n"); printf(msg); printf("\n");
   exit(1);
   }
   
   
