#include <iostream.h>
#include <math.h>
#include "global.h"

void timestep(Matrix<real>* W, Matrix<real>& x, Matrix<real>& dt)
   {
   int i,j;
   
   for(i=2;i<=imp1;i++)
      for(j=2;j<=jmp1;j++)
         dt(i,j)=CFL*(x(i,j)-x(i-1,j))
                 /(fabs(W[1](i,j)/W[0](i,j))+sqrt(gam*W[4](i,j)/W[0](i,j)));
   }
   

real Mach(Matrix<real>* W,int i,int j)
   {
   return sqrt((W[1](i,j)*W[1](i,j)+W[2](i,j)*W[2](i,j))
              /(W[4](i,j)*W[0](i,j)*gam));
   }           

void flux_ij(Matrix<real>* W, int i, int j,real f[4])
   {
   real u=0.5*(W[1](i,j)/W[0](i,j)+W[1](i+1,j)/W[0](i+1,j));
   
   f[0]=0.5*u*(W[0](i,j)+W[0](i+1,j));
   f[1]=0.5*(u*(W[1](i,j)+W[1](i+1,j))+W[4](i,j)+W[4](i+1,j));
   f[2]=0.;
   f[3]=0.5*u*(W[3](i,j)+W[3](i+1,j)+W[4](i,j)+W[4](i+1,j));
   }

void boundary(Matrix<real>* W, Matrix<real> flux1[2][4], Matrix<real> flux2[2][4], Matrix<real>& vol)
   {
   int i,j;
   real M,MM,r1,r2,r3,a0,a1,a2,rad;
   real fp[4],f[4];

   
   for(j=2;j<=jmp1;j++)
      {
      r2=W[1](1,j)/W[0](1,j)+2.*sqrt(gam*W[4](1,j)/W[0](1,j))/(gam-1.);
      r3=W[1](2,j)/W[0](2,j)-2.*sqrt(gam*W[4](2,j)/W[0](2,j))/(gam-1.);

      M=2./(gam-1.)*(r2+r3)/(r2-r3);
      if(M<0 || M>1) 
         M=Mach(W,2,j);
      
      MM=1.+0.5*(gam-1.)*M*M;
      
      W[4](1,j)=P0i/pow(MM,gam/(gam-1.));
      W[0](1,j)=W[4](1,j)/((cp-cv)*T0i)*MM;
      W[1](1,j)=W[0](1,j)*M*sqrt(gam*W[4](1,j)/W[0](1,j));
      W[2](1,j)=0.;
      W[3](1,j)=W[4](1,j)/(gam-1.)+0.5*(W[1](1,j)*W[1](1,j)
                                       +W[2](1,j)*W[2](1,j))
                                       /W[0](1,j);


      r1=W[4](imp1,j)/pow(W[0](imp1,j),gam);
      r2=W[1](imp1,j)/W[0](imp1,j)+2.*sqrt(gam*W[4](imp1,j)/W[0](imp1,j))/(gam-1.);
      
      W[4](imp2,j)=pe;
      W[0](imp2,j)=pow(W[4](imp2,j)/r1,1./gam);
      W[1](imp2,j)=W[0](imp2,j)*(r2-2.*sqrt(gam*W[4](imp2,j)/W[0](imp2,j))/(gam-1.));
      
      
      W[2](imp2,j)=0.;
      W[3](imp2,j)=W[4](imp2,j)/(gam-1.)+0.5*(W[1](imp2,j)*W[1](imp2,j)
                                             +W[2](imp2,j)*W[2](imp2,j))
                                             /W[0](imp2,j);

      flux_ij(W,1,j,f);
      
      flux1[0][0](1,j)=f[0];
      flux1[0][1](1,j)=f[1];
      flux1[0][2](1,j)=f[2];
      flux1[0][3](1,j)=f[3];                  

      flux_ij(W,imp1,j,f);
      
      flux1[0][0](imp1,j)=f[0];
      flux1[0][1](imp1,j)=f[1];
      flux1[0][2](imp1,j)=f[2];
      flux1[0][3](imp1,j)=f[3]; 
                       
      }
   }

void Euler_flux(Matrix<real>* W, Matrix<real> flux1[2][4], Matrix<real> flux2[2][4])
   {
   int i,j;
   real fp[4],f[4];
   real dp,dpp,u;
   
   for(i=2;i<=imp1;i++)
      for(j=2;j<=jmp1;j++)
         {
         W[4](i,j)=(gam-1.)*(W[3](i,j)
                       -0.5*(W[1](i,j)*W[1](i,j)
                            +W[2](i,j)*W[2](i,j))
                            /W[0](i,j));
         
         if(W[4](i,j)<0) cout<<"negative pressure !!!!!!!"<<endl;
         }
   
   for(i=2;i<imp1;i++)
      for(j=2;j<=jmp1;j++)
         {
         flux_ij(W,i,j,f);
         
         flux1[0][0](i,j)=f[0];
         flux1[0][1](i,j)=f[1];
         flux1[0][2](i,j)=f[2];
         flux1[0][3](i,j)=f[3];                  
         }
   }
   
void diff_flux(Matrix<real>* W, Matrix<real>* dflux1, Matrix<real>* dflux2)
   {
   int i,j;
   real eps2,eps4,dpp,dp;
   real d,u,E,dflx0,dflx1,dflx3;
   real r[4],f[4],fp[4];
   
   for(j=2;j<=jmp1;j++)
      {
      dpp=fabs((W[4](3,j)-2*W[4](2,j)+W[4](1,j))
              /(W[4](3,j)+2*W[4](2,j)+W[4](1,j)));
      for(i=2;i<imp1;i++)
         {
         dp=dpp;
         
         dpp=fabs((W[4](i+2,j)-2*W[4](i+1,j)+W[4](i,j))
                 /(W[4](i+2,j)+2*W[4](i+1,j)+W[4](i,j)));

         
         r[0]=fabs((W[1](i,j)+W[1](i+1,j))/(W[0](i,j)+W[0](i+1,j)))
              +0.5*(sqrt(gam*W[4](i,j)/W[0](i,j))+sqrt(gam*W[4](i+1,j)/W[0](i+1,j)));
         r[0]=r[1]=r[2]=r[3]=0.5*r[0];
              
         eps2=kap2*max(dpp,dp); 
         eps4=max(0,kap4-eps2);
         
         dflux1[0](i,j) =r[0]*( eps2*(W[0](i+1,j)                  -W[0](i,j))
                               -eps4*(W[0](i+2,j)-3.*W[0](i+1,j)+3.*W[0](i,j)-W[0](i-1,j)));
                       
         dflux1[1](i,j) =r[1]*( eps2*(W[1](i+1,j)                  -W[1](i,j))
                               -eps4*(W[1](i+2,j)-3.*W[1](i+1,j)+3.*W[1](i,j)-W[1](i-1,j)));
                       
         dflux1[2](i,j) =r[2]*( eps2*(W[2](i+1,j)                  -W[2](i,j))
                               -eps4*(W[2](i+2,j)-3.*W[2](i+1,j)+3.*W[2](i,j)-W[2](i-1,j)));
                       
         dflux1[3](i,j) =r[3]*( eps2*(W[3](i+1,j)                  -W[3](i,j))
                               -eps4*(W[3](i+2,j)-3.*W[3](i+1,j)+3.*W[3](i,j)-W[3](i-1,j)));
         
         }                      
      }   
   }
   
void residual(Matrix<real>& vol, Matrix<real>*  surf1, Matrix<real>*  surf2,
                           Matrix<real> flux1[2][4], Matrix<real> flux2[2][4], 
                           Matrix<real>* dflux1, Matrix<real>* dflux2, Matrix<real>* R,Matrix<real>* W)
   {
   int i,j;
   
   for(i=2;i<=imp1;i++)
      for(j=2;j<=jmp1;j++)
         {
         
         //                        x1-Direction                           x2-Direction 
         
         R[0](i,j) = (flux1[0][0](i  ,j  )*surf1[0](i  ,j  ) // +flux2[0][0](i  ,j  )*surf2[0](i  ,j  )  // outlet1
                     +flux1[1][0](i  ,j  )*surf1[1](i  ,j  ) // +flux2[1][0](i  ,j  )*surf2[1](i  ,j  )  // outlet2
                     
                     -flux1[0][0](i-1,j  )*surf1[0](i-1,j  ) // -flux2[0][0](i-1,j  )*surf2[0](i-1,j  )  // inlet1
                     -flux1[1][0](i  ,j-1)*surf1[1](i  ,j-1) // -flux2[1][0](i  ,j-1)*surf2[1](i  ,j-1)  // inlet2
                     
                     -dflux1  [0](i  ,j  )                   // -dflux2  [0](i  ,j  )                   
                     +dflux1  [0](i-1,j  )                   // +dflux2  [0](i  ,j-1)
                     )/vol(i,j);
   


         R[1](i,j) = (flux1[0][1](i  ,j  )*surf1[0](i  ,j  ) // +flux2[0][1](i  ,j  )*surf2[0](i  ,j  ) 
                     +flux1[1][1](i  ,j  )*surf1[1](i  ,j  ) // +flux2[1][1](i  ,j  )*surf2[1](i  ,j  ) 
 
                     -flux1[0][1](i-1,j  )*surf1[0](i-1,j  ) // -flux2[0][1](i-1,j  )*surf2[0](i-1,j  ) 
                     -flux1[1][1](i  ,j-1)*surf1[1](i  ,j-1) // -flux2[1][1](i  ,j-1)*surf2[1](i  ,j-1) 
 
                     -dflux1  [1](i  ,j  )                   // -dflux2  [1](i  ,j  )                   
                     +dflux1  [1](i-1,j  )                   // +dflux2  [1](i  ,j-1)                   
                     
                     -W[4](i,j)*(surf1[0](i,j)-surf1[0](i-1,j))
                     
                     )/vol(i,j);



         R[2](i,j) = (flux1[0][2](i  ,j  )*surf1[0](i  ,j  ) // +flux2[0][2](i  ,j  )*surf2[0](i  ,j  ) 
                     +flux1[1][2](i  ,j  )*surf1[1](i  ,j  ) // +flux2[1][2](i  ,j  )*surf2[1](i  ,j  ) 
 
                     -flux1[0][2](i-1,j  )*surf1[0](i-1,j  ) // -flux2[0][2](i-1,j  )*surf2[0](i-1,j  ) 
                     -flux1[1][2](i  ,j-1)*surf1[1](i  ,j-1) // -flux2[1][2](i  ,j-1)*surf2[1](i  ,j-1) 
 
                     -dflux1  [2](i  ,j  )                   // -dflux2  [2](i  ,j  )                   
                     +dflux1  [2](i-1,j  )                   // +dflux2  [2](i  ,j-1)
                     )/vol(i,j);



         R[3](i,j) = (flux1[0][3](i  ,j  )*surf1[0](i  ,j  ) // +flux2[0][3](i  ,j  )*surf2[0](i  ,j  ) 
                     +flux1[1][3](i  ,j  )*surf1[1](i  ,j  ) // +flux2[1][3](i  ,j  )*surf2[1](i  ,j  ) 
 
                     -flux1[0][3](i-1,j  )*surf1[0](i-1,j  ) // -flux2[0][3](i-1,j  )*surf2[0](i-1,j  ) 
                     -flux1[1][3](i  ,j-1)*surf1[1](i  ,j-1) // -flux2[1][3](i  ,j-1)*surf2[1](i  ,j-1) 
 
                     -dflux1  [2](i  ,j  )                   // -dflux2  [3](i  ,j  )                   
                     +dflux1  [2](i-1,j  )                   // +dflux2  [3](i  ,j-1)
                     )/vol(i,j);
         }             
   }
   
void march(Matrix<real>* W,   Matrix<real>*   Wold, 
           Matrix<real>& vol, Matrix<real>*  surf1, Matrix<real>*  surf2,
                        Matrix<real> flux1[2][4], Matrix<real> flux2[2][4], 
                        Matrix<real>* dflux1, Matrix<real>* dflux2, Matrix<real>* R, Matrix<real>& dt)                             
   {
   real a[]={0.25, 1./3., 0.5, 1.0};
   int m,i,nix;
   

   for(m=0;m<4;m++)
      {
      boundary(W,  flux1,  flux2, vol);
      Euler_flux(W,  flux1,  flux2);
      diff_flux(W, dflux1, dflux2);
      residual(vol,  surf1,  surf2, flux1, flux2, dflux1, dflux2, R,W);
      
      if(!m)
         {
         Wold[0]=W[0];
         Wold[1]=W[1];
         Wold[2]=W[2];
         Wold[3]=W[3];
         }

      W[0]=Wold[0]-a[m]*dt*R[0];
      W[1]=Wold[1]-a[m]*dt*R[1];
      W[2]=Wold[2]-a[m]*dt*R[2];
      W[3]=Wold[3]-a[m]*dt*R[3];
      }
   }
                                       











